local ADDONNAME, ns = ...
ns = ns or {}

-- create addon
local Addon = LibStub("AceAddon-3.0"):NewAddon(ADDONNAME, "AceConsole-3.0", "AceEvent-3.0", "AceTimer-3.0")

-- load libs
local LSM = LibStub("LibSharedMedia-3.0")
local LSK = LibStub("LibSink-2.0")

-- load locale
local L = LibStub("AceLocale-3.0"):GetLocale(ADDONNAME)

-- WoW APIs
local _G = _G
local C_Vignettes = C_Vignettes

-- Lua API
local pairs, tostring, next, type, tonumber, unpack = pairs, tostring, next, type, tonumber, unpack
local rawget, rawset, wipe = rawget, rawset, wipe

-- GLOBALS: LibStub, MVAnnouncerDefaults

function Addon:OnInitialize()
    if type(MVAnnouncerDefaults) ~= "table" then
        MVAnnouncerDefaults = {}
    end
    if MVAnnouncerDefaults.seen == nil then
        MVAnnouncerDefaults.seen = false
    end
    if MVAnnouncerDefaults.active == nil then
        MVAnnouncerDefaults.active = true
    end
    if MVAnnouncerDefaults.showIconBefore == nil then
        MVAnnouncerDefaults.showIconBefore = true
    end
    if MVAnnouncerDefaults.showIconAfter == nil then
        MVAnnouncerDefaults.showIconAfter = false
    end
    if MVAnnouncerDefaults.mainTextColour == nil then
        MVAnnouncerDefaults.mainTextColour = "FFFFFFFF"
    end
    if MVAnnouncerDefaults.objectTextColour == nil then
        MVAnnouncerDefaults.objectTextColour = "FFB58900"
    end
    if MVAnnouncerDefaults.iconSize == nil then
        MVAnnouncerDefaults.iconSize = 16
    end
    if MVAnnouncerDefaults.playSound == nil then
        MVAnnouncerDefaults.playSound = true
    end
    if MVAnnouncerDefaults.sound == nil then
        MVAnnouncerDefaults.sound = ns.defaultSoundName
    end
    if type(MVAnnouncerDefaults.sinks) ~= "table" then
        MVAnnouncerDefaults.sinks = {}
    end
    if MVAnnouncerDefaults.sinks['**'] == nil then
        MVAnnouncerDefaults.sinks['**'] = {
            sink20OutputSink = "RaidWarning",
        }

        -- First time we load this, so add a default sink
        MVAnnouncerDefaults.sinks.BuildInDefaultOutputSink = {
            sink20OutputSink = "RaidWarning",
        }
    end

    ns.defaults.profile.objectIconTypeOfInterest['**']['**'] = MVAnnouncerDefaults
    self.db = LibStub("AceDB-3.0"):New("MVAnnouncerDB", ns.defaults, true)
    self.db.RegisterCallback(self, "OnProfileCopied", "OnDBUpdate")
    self.db.RegisterCallback(self, "OnDatabaseReset", "OnDBUpdate")
    self.db.RegisterCallback(self, "OnNewProfile", "OnDBUpdate")
    self.db.RegisterCallback(self, "OnProfileReset", "OnDBUpdate")
    self.db.RegisterCallback(self, "OnProfileChanged", "OnDBUpdate")
    ns.db = self.db

    self:RegisterChatCommand("minimapvignetteannouncer", "CMD")
    self:RegisterChatCommand("mvannouncer", "CMD")
    self:RegisterChatCommand("mva", "CMD")
    if L["CHAT_COMMAND"] ~= "CHAT_COMMAND" then
        self:RegisterChatCommand(L["CHAT_COMMAND"], "CMD")
    end

    self:RegisterOptions()
end

function Addon:OnDBUpdate(event, db, profile)
    self:CleanOptions()
end

function Addon:OnEnable()
    self:RegisterEvent("VIGNETTE_ADDED")
    self:EnableScan()
end

function Addon:OnDisable()
    self:UnregisterEvent("VIGNETTE_ADDED")
    self:DisableScan()
end

do -- Scanning
    -- Use that only when a vignette on the map with GetNumVignettes() is found
    -- but does not trigger VIGNETTE_ADDED.
    -- enable with: /run LibStub("AceAddon-3.0"):GetAddon("MVAnnouncer").db.profile.scan.active = true
    -- disable with: /run LibStub("AceAddon-3.0"):GetAddon("MVAnnouncer").db.profile.scan.active = false
    -- Set scan interval(seconds) with: /run LibStub("AceAddon-3.0"):GetAddon("MVAnnouncer").db.profile.scan.interval = 10
    local scanTimer

    function Addon:EnableScan()
        if self.db.profile.scan.active then
            self:DisableScan()
            scanTimer = self:ScheduleRepeatingTimer("Scan", self.db.profile.scan.interval)
        else
            self:DisableScan()
        end
    end
    function Addon:DisableScan()
        self:CancelTimer(scanTimer)
    end
    function Addon:Scan()
        local numVignettes = C_Vignettes.GetNumVignettes()
        for i=1, numVignettes do
            local vignetteGUID = C_Vignettes.GetVignetteGUID(i)
            local ofsX, ofsY, name, objectIcon = C_Vignettes.GetVignetteInfoFromInstanceID(vignetteGUID)
            if objectIcon and self.db.profile.scan.objects[tostring(objectIcon)] then
                self:VIGNETTE_ADDED("VIGNETTE_ADDED", vignetteGUID)
            end
        end
    end
end

do
    local cooldown = {}

    function Addon:PlayControlledSoundFile(soundFile, channel)
        if ( cooldown[soundFile] or 0 ) <= _G.GetTime() then
            cooldown[soundFile] = _G.GetTime() + self.db.profile.soundCooldown
            _G.PlaySoundFile(soundFile, channel)
        end
    end
end

function Addon:Announce(vignetteGUID)
    if not vignetteGUID then
        return
    end

    local ofsX, ofsY, name, objectIcon = C_Vignettes.GetVignetteInfoFromInstanceID(vignetteGUID)

    if not objectIcon then
        objectIcon = L["Unknown Icon"]
    else
        objectIcon = tostring(objectIcon)
    end
    if not name then
        name = L["Unknown vignette"]
    else
        name = tostring(name)
    end

    if not self.db.profile.objectIconTypeOfInterest[objectIcon][name].seen then
        self.db.profile.objectIconTypeOfInterest[objectIcon][name].seen = true
        --for key in pairs(self.db.profile.objectIconTypeOfInterest[objectIcon][name].sinks) do
        --    self.db.profile.objectIconTypeOfInterest[objectIcon][name].sinks[key].touched = true
        --end
    end

    if self.db.profile.objectIconTypeOfInterest[objectIcon].active and self.db.profile.objectIconTypeOfInterest[objectIcon][name].active then
        local msg = ""

        local left, right, top, bottom
        if self.db.profile.objectIconTypeOfInterest[objectIcon][name].showIconBefore and tonumber(objectIcon) then
            if not left then
                left, right, top, bottom = unpack(ns.OBJECTICONS_coords[objectIcon])
                left, right, top, bottom = left*ns.objectIconTextureSize, right*ns.objectIconTextureSize, top*ns.objectIconTextureSize, bottom*ns.objectIconTextureSize
            end

            msg = msg..ns.OBJECTICONS_chatTexture:format(self.db.profile.objectIconTypeOfInterest[objectIcon][name].iconSize, self.db.profile.objectIconTypeOfInterest[objectIcon][name].iconSize, left, right, top, bottom)
        end

        local mainTextColour = self.db.profile.objectIconTypeOfInterest[objectIcon][name].mainTextColour
        local objectTextColour = self.db.profile.objectIconTypeOfInterest[objectIcon][name].objectTextColour

        msg = msg..L["|c%sFound |c%s%q|c%s look on your minimap!|r"]:format(mainTextColour, objectTextColour, name, mainTextColour)

        if self.db.profile.objectIconTypeOfInterest[objectIcon][name].showIconAfter and tonumber(objectIcon) then
            if not left then
                left, right, top, bottom = unpack(ns.OBJECTICONS_coords[objectIcon])
                left, right, top, bottom = left*ns.objectIconTextureSize, right*ns.objectIconTextureSize, top*ns.objectIconTextureSize, bottom*ns.objectIconTextureSize
            end

            msg = msg..ns.OBJECTICONS_chatTexture:format(self.db.profile.objectIconTypeOfInterest[objectIcon][name].iconSize, self.db.profile.objectIconTypeOfInterest[objectIcon][name].iconSize, left, right, top, bottom)
        end

        if self.db.profile.objectIconTypeOfInterest[objectIcon][name].playSound then
            self:PlayControlledSoundFile(LSM:Fetch(LSM.MediaType.SOUND, self.db.profile.objectIconTypeOfInterest[objectIcon][name].sound), "Master")
        end

        for key in pairs(self.db.profile.objectIconTypeOfInterest[objectIcon][name].sinks) do
            if _G.CHAT_TIMESTAMP_FORMAT and self.db.profile.objectIconTypeOfInterest[objectIcon][name].sinks[key].sink20OutputSink == "ChatFrame" then
                ns.sinkTable[objectIcon..ns.SPLITKEY..name..ns.SPLITKEY..key]:Pour(_G.BetterDate(_G.CHAT_TIMESTAMP_FORMAT, _G.time())..msg)
            else
                ns.sinkTable[objectIcon..ns.SPLITKEY..name..ns.SPLITKEY..key]:Pour(msg)
            end
        end
    end
    --[===[@debug@
    -- lets cache Vignette GUID for testing,
    tinsert(self.db.global.vignette_cache[objectIcon][name], vignetteGUID)
    self:Printf("Vignette: %s", vignetteGUID)
    --@end-debug@]===]
end

function Addon:VIGNETTE_ADDED(event, vignetteGUID)
    --[[
    local ofsX, ofsY, name, objectIcon = C_Vignettes.GetVignetteInfoFromInstanceID(vignetteGUID)
    if not objectIcon then
        -- todo: find way to get name and objectIcon
    end
    ]]
    if ns.foundCooldowns[vignetteGUID] <= _G.GetTime() then
        ns.foundCooldowns[vignetteGUID] = _G.GetTime() + self.db.profile.cooldownTime
        self:Announce(vignetteGUID)
    end
end

-- LOCALE: L["CONSOLE_HELP"] = "Chat commands:\n|cFF278BCB/mva|r - opens settings\n|cFF278BCB/mva|r manual - add a vignette manual to the settings\n|cFF278BCB/mva|r remove - remove a vignette from the settings"
-- LOCALE: L["CONSOLE_HELP_REMOVE"] = '|cFF278BCB/mva|r remove objecIconID objectName\nObjectName must be the same as in the option shown.\n\ne.g: 40 "Ohter Chest"'
function Addon:CMD(input)
    local arg1, lastPos = self:GetArgs(input, 1)
    if not arg1 then
        self:OpenSettings()
    elseif arg1:lower() == "manual" then
        self:CMD_Manual(input:sub(lastPos))
    elseif arg1:lower() ==  "remove" then
        local objectIcon, objectName, lastPos = self:GetArgs(input, 2, lastPos)
        if tonumber(objectIcon) and type(objectName) == "string" and objectName:len() > 0 and lastPos == 1e9 then
            self.db.profile.objectIconTypeOfInterest[objectIcon][objectName] = nil
            self:RemoveVignette(objectIcon, objectName)
        else
            self:Print(L["CONSOLE_HELP_REMOVE"])
        end
    else
        self:Print(L["CONSOLE_HELP"])
    end
end

-- LOCALE: L["CONSOLE_HELP_MANUAL"] = '|cFF278BCB/mva|r manual objecIconID objectName active playSound\nObjectName must be the same as in the tooltip of the minimap vignette, use quotes for names with spacing.\n\nActive chest icon without sound e.g: 40 Chest true false\nInactive skull icon with default sound settings e.g: 41 Rar false\nChest icon with default active setting and active sound e.g: 40 "Other Chest" nil true'
function Addon:CMD_Manual(input)
    local objectIcon, objectName, lastPos = self:GetArgs(input, 2)
    if tonumber(objectIcon) and type(objectName) == "string" and objectName:len() > 0 then
        if lastPos == 1e9 then
            self.db.profile.objectIconTypeOfInterest[objectIcon][objectName].seen = true
        else
            local isOK = true
            local active, lastPos = self:GetArgs(input, 1, lastPos)

            active, isOK = ns.GetBooleanValue(active)
            if active ~= nil and isOK then
                self.db.profile.objectIconTypeOfInterest[objectIcon][objectName].active = active
            elseif not isOK then
                self:Print(L["CONSOLE_HELP_MANUAL"])
                return
            end

            if lastPos ~= 1e9 then
                local playSound, lastPos = self:GetArgs(input, 1, lastPos)
                playSound, isOK = ns.GetBooleanValue(playSound)
                if playSound ~= nil and isOK then
                    self.db.profile.objectIconTypeOfInterest[objectIcon][objectName].playSound = playSound
                elseif not isOK then
                    self:Print(L["CONSOLE_HELP_MANUAL"])
                    return
                end
            end

            self.db.profile.objectIconTypeOfInterest[objectIcon][objectName].seen = true
        end
        self:NotifyOptionsChange()
    else
        self:Print(L["CONSOLE_HELP_MANUAL"])
    end
end

function Addon:OpenSettings(...)
    _G.InterfaceOptionsFrame_OpenToCategory(ADDONNAME)
    if not _G.InterfaceOptionsFrameAddOns:IsShown() then
        _G.InterfaceOptionsFrame_OpenToCategory(ADDONNAME)
    end
end

-- options
do
    local options = {}
    local defaultsOptions = {}
    local announcementOptions = {}
    local outputMinChar, outputMaxChar = 3, 32

    local OptionsMethods = {}
    function OptionsMethods:Set(info, value, ...)
        Addon.db.profile[info[#info]] = value
    end
    function OptionsMethods:Get(info, value, ...)
        return Addon.db.profile[info[#info]]
    end
    local tmpDefaults = {
        sinks = {}
    }
    local tmpDefaultsSinksTable = setmetatable({}, {__index=function(t,k)
        local v = rawget(t, k)

        if type(v) ~= "table" then
            v = LSK:Embed({})
            tmpDefaults.sinks[k] = tmpDefaults.sinks[k] or {}
            v:SetSinkStorage(tmpDefaults.sinks[k])
            rawset(t, k, v)
        end

        return v
    end})
    function OptionsMethods:SetDefaults(info, value, ...)
        tmpDefaults[info[#info]] = value
    end
    function OptionsMethods:GetDefaults(info, value, ...)
        if tmpDefaults[info[#info]] ~= nil then
            return tmpDefaults[info[#info]]
        else
            return MVAnnouncerDefaults[info[#info]]
        end
    end
    function OptionsMethods:ApplyDefaults(info, value, ...)
        local db = Addon.db
        db:RegisterDefaults(nil)

        for key, value in pairs(tmpDefaults) do
            if key == "sinks" then
                for sinkKey, sinkValue in pairs(value) do
                    if type(sinkValue) == "table" then
                        for settingKey, settingValue in pairs(sinkValue) do
                            if not MVAnnouncerDefaults[key][sinkKey] then
                                MVAnnouncerDefaults[key][sinkKey] = {}
                            end
                            MVAnnouncerDefaults[key][sinkKey][settingKey] = settingValue
                        end
                    else
                        MVAnnouncerDefaults[key][sinkKey] = nil
                    end
                end
            else
                MVAnnouncerDefaults[key] = value
            end
        end

        ns.defaults.profile.objectIconTypeOfInterest['**']['**'] = MVAnnouncerDefaults
        db:RegisterDefaults(ns.defaults)

        Addon:CleanOptions()
    end
    function OptionsMethods:DeleteDefaultSink(info, value, ...)
        local sinkName = info.arg

        tmpDefaults.sinks[sinkName] = false
        tmpDefaultsSinksTable[sinkName] = nil
        defaultsOptions.args[sinkName.."_sink"] = nil
    end
    function OptionsMethods:ValidateDefaultOutputName(info, value, ...)
        value = value:trim()
        if value:match("[%w]+[%w%s]*[%w]+") and value:len() >= outputMinChar and value:len() <= outputMaxChar then
            if not tmpDefaults.sinks[value] then
                return true
            end

            return L["Output %q exits already!"]:format(value)
        end

        return L["Output name length may be between %d and %d character and consist of letters, numbers, and spacings."]:format(outputMinChar, outputMaxChar)
    end
    function OptionsMethods:CreateNewDefaultOutput(info, value, ...)
        value = value:trim()
        tmpDefaults.sinks[value] = tmpDefaults.sinks[value] or {}
        tmpDefaults.sinks[value].sink20OutputSink = "Default"
    end
    function OptionsMethods:SetAnnoucementsOptions(info, value, ...)
        local objectIcon = ns.SPLITKEY:split(info.arg)

        Addon.db.profile.objectIconTypeOfInterest[objectIcon][info[#info]] = value
    end
    function OptionsMethods:GetAnnoucementsOptions(info, value, ...)
        local objectIcon = ns.SPLITKEY:split(info.arg)

        return Addon.db.profile.objectIconTypeOfInterest[objectIcon][info[#info]]
    end
    function OptionsMethods:SetAnnoucementSettings(info, value, ...)
        local objectIcon, name = ns.SPLITKEY:split(info.arg)

        Addon.db.profile.objectIconTypeOfInterest[objectIcon][name][info[#info]] = value
    end
    function OptionsMethods:GetAnnoucementSettings(info, value, ...)
        local objectIcon, name = ns.SPLITKEY:split(info.arg)

        return Addon.db.profile.objectIconTypeOfInterest[objectIcon][name][info[#info]]
    end
    function OptionsMethods:ValidateOutputName(info, value, ...)
        value = value:trim()
        if value:match("[%w]+[%w%s]*[%w]+") and value:len() >= outputMinChar and value:len() <= outputMaxChar then
            local objectIcon, name = ns.SPLITKEY:split(info.arg)
            if Addon.db.profile.objectIconTypeOfInterest[objectIcon][name].sinks[value].touched == nil then
                return true
            end

            return L["Output %q exits already!"]:format(value)
        end

        return L["Output name length may be between %d and %d character and consist of letters, numbers, and spacings."]:format(outputMinChar, outputMaxChar)
    end
    function OptionsMethods:CreateNewOutput(info, value, ...)
        value = value:trim()
        local objectIcon, name = ns.SPLITKEY:split(info.arg)
        Addon.db.profile.objectIconTypeOfInterest[objectIcon][name].sinks[value].touched = true
    end
    function OptionsMethods:GetNewOutput(info, value, ...)
        return ""
    end
    function OptionsMethods:DeleteSink(info, value, ...)
        local objectIcon, name, sinkName = ns.SPLITKEY:split(info.arg)

        announcementOptions[objectIcon..name].args[sinkName] = nil
        Addon.db.profile.objectIconTypeOfInterest[objectIcon][name].sinks[sinkName:sub(1, -6)] = false
        ns.sinkTable[info.arg] = nil
    end
    function OptionsMethods:SetColour(info, r, g, b, a, ...)
        self:SetAnnoucementSettings(info, ns.RGBAToHex(r,g,b,a), ...)
    end
    function OptionsMethods:GetColour(info, value, ...)
        return ns.HexToRGBA(self:GetAnnoucementSettings(info, value, ...))
    end
    function OptionsMethods:SetDefaultColour(info, r, g, b, a, ...)
        self:SetDefaults(info, ns.RGBAToHex(r,g,b,a), ...)
    end
    function OptionsMethods:GetDefaultColour(info, value, ...)
        return ns.HexToRGBA(self:GetDefaults(info, value, ...))
    end

    function Addon.GetOptionsTable(uiType, uiName, appName)
        local self = Addon
        options.name = ADDONNAME
        options.type = "group"
        options.childGroups = "tab"
        options.inline = true
        options.handler = OptionsMethods
        options.set = "Set"
        options.get = "Get"
        options.args = options.args or {}

        options.args.active = options.args.active or {}
        options.args.active.name = L["Active"]
        options.args.active.type = "toggle"
        options.args.active.desc = L["Enable/Disable %s on this profile."]:format(ADDONNAME)
        options.args.active.order = 0

        options.args.cooldownTime = options.args.cooldownTime or {}
        options.args.cooldownTime.name = L["Warning Cooldown"]
        options.args.cooldownTime.type = "range"
        options.args.cooldownTime.desc = L["How long should be wait before the same vignette can trigger a new announcement.\n%g Seconds."]:format(self.db.profile.cooldownTime)
        options.args.cooldownTime.min = 0
        options.args.cooldownTime.max = 900
        options.args.cooldownTime.step = 0.01
        options.args.cooldownTime.bigStep = 1
        options.args.cooldownTime.softMax = 60
        options.args.cooldownTime.order = 0.5

        options.args.soundCooldown = options.args.soundCooldown or {}
        options.args.soundCooldown.name = L["Sound Cooldown"]
        options.args.soundCooldown.type = "range"
        options.args.soundCooldown.desc = L["Time until the same sound can be played again."]
        options.args.soundCooldown.min = 0
        options.args.soundCooldown.max = 900
        options.args.soundCooldown.step = 0.01
        options.args.soundCooldown.bigStep = 1
        options.args.soundCooldown.softMax = 10
        options.args.soundCooldown.order = 0.55

        for objectIcon, announcements in pairs(Addon.db.profile.objectIconTypeOfInterest) do
            options.args[objectIcon] = options.args[objectIcon] or {}

            if tonumber(objectIcon) then
                options.args[objectIcon].icon = ns.OBJECTICONS_texture
                options.args[objectIcon].iconCoords = ns.OBJECTICONS_coords[objectIcon]
                options.args[objectIcon].name = L["Vignette: %s"]:format(objectIcon)
            else
                options.args[objectIcon].name = objectIcon
            end

            options.args[objectIcon].type = "group"
            options.args[objectIcon].childGroups = "tree"
            options.args[objectIcon].inline = false
            options.args[objectIcon].set = "SetAnnoucementsOptions"
            options.args[objectIcon].get = "GetAnnoucementsOptions"
            options.args[objectIcon].args = options.args[objectIcon].args or {}

            if tonumber(objectIcon) then
                options.args[objectIcon].args.image = options.args[objectIcon].args.image or {}
                options.args[objectIcon].args.image.name = L["-  \"Vignette\""]
                options.args[objectIcon].args.image.type = "description"
                options.args[objectIcon].args.image.image = ns.OBJECTICONS_texture
                options.args[objectIcon].args.image.imageCoords = ns.OBJECTICONS_coords[objectIcon]
                options.args[objectIcon].args.image.fontSize = "medium"
                options.args[objectIcon].args.image.order = 0
            end

            options.args[objectIcon].args.active = options.args[objectIcon].args.active or {}
            options.args[objectIcon].args.active.name = L["Active"]
            options.args[objectIcon].args.active.desc = L["If this vignette type active."]
            options.args[objectIcon].args.active.arg = objectIcon
            options.args[objectIcon].args.active.type = "toggle"
            options.args[objectIcon].args.active.order = 0.1

            for name, value in pairs(announcements) do
                if type(value) == "table" then
                    options.args[objectIcon].args[name] = Addon.GetAnnoucementSettings(uiType, uiName, appName, objectIcon, name)
                end
            end
        end

        return options
    end

    function Addon.GetAnnoucementSettings(uiType, uiName, appName, objectIcon, name)
        announcementOptions[objectIcon..name] = announcementOptions[objectIcon..name] or {}
        local settings = announcementOptions[objectIcon..name]
        local arg = objectIcon..ns.SPLITKEY..name

        settings.name = name
        settings.type = "group"
        settings.set = "SetAnnoucementSettings"
        settings.get = "GetAnnoucementSettings"
        settings.args = settings.args or {}

        settings.args.active = settings.args.active or {}
        settings.args.active.name = L["Active"]
        settings.args.active.type = "toggle"
        settings.args.active.desc = L["If this announcement active."]
        settings.args.active.arg = arg
        settings.args.active.width = "double"
        settings.args.active.order = 0

        settings.args.iconSize = settings.args.iconSize or {}
        settings.args.iconSize.name = L["Icon size"]
        settings.args.iconSize.type = "range"
        settings.args.iconSize.desc = L["Size of the vignette icon."]
        settings.args.iconSize.arg = arg
        settings.args.iconSize.min = 8
        settings.args.iconSize.max = 512
        settings.args.iconSize.step = 1
        settings.args.iconSize.bigStep = 4
        settings.args.iconSize.softMax = 64
        settings.args.iconSize.order = 2

        settings.args.showIconBefore = settings.args.showIconBefore or {}
        settings.args.showIconBefore.name = L["Show an icon before the announcement"]
        settings.args.showIconBefore.type = "toggle"
        settings.args.showIconBefore.desc = L["Should be shown an icon before the announcement."]
        settings.args.showIconBefore.arg = arg
        settings.args.showIconBefore.width = "full"
        settings.args.showIconBefore.order = 3.1

        settings.args.showIconAfter = settings.args.showIconAfter or {}
        settings.args.showIconAfter.name = L["Show an icon after the announcement"]
        settings.args.showIconAfter.type = "toggle"
        settings.args.showIconAfter.desc = L["Should be shown an icon after the announcement."]
        settings.args.showIconAfter.arg = arg
        settings.args.showIconAfter.width = "full"
        settings.args.showIconAfter.order = 3.2

        settings.args.mainTextColour = settings.args.mainTextColour or {}
        settings.args.mainTextColour.name = L["Main text colour"]
        settings.args.mainTextColour.type = "color"
        settings.args.mainTextColour.desc = L["Main text colour of the announcement"]
        settings.args.mainTextColour.hasAlpha = true
        settings.args.mainTextColour.arg = arg
        settings.args.mainTextColour.set = "SetColour"
        settings.args.mainTextColour.get = "GetColour"
        settings.args.mainTextColour.width = "full"
        settings.args.mainTextColour.order = 4.1

        settings.args.objectTextColour = settings.args.objectTextColour or {}
        settings.args.objectTextColour.name = L["Vignette text colour"]
        settings.args.objectTextColour.type = "color"
        settings.args.objectTextColour.desc = L["Vignette text colour of the announcement"]
        settings.args.objectTextColour.hasAlpha = true
        settings.args.objectTextColour.arg = arg
        settings.args.objectTextColour.set = "SetColour"
        settings.args.objectTextColour.get = "GetColour"
        settings.args.objectTextColour.width = "full"
        settings.args.objectTextColour.order = 4.2

        settings.args.playSound = settings.args.playSound or {}
        settings.args.playSound.name = L["Play sound"]
        settings.args.playSound.type = "toggle"
        settings.args.playSound.desc = L["Should this announcement play a sound."]
        settings.args.playSound.arg = arg
        settings.args.playSound.order = 5.1

        settings.args.sound = settings.args.sound or {}
        settings.args.sound.name = L["Sound"]
        settings.args.sound.type = "select"
        settings.args.sound.desc = L["Which sound should be played."]
        settings.args.sound.arg = arg
        settings.args.sound.values = LSM:HashTable(LSM.MediaType.SOUND)
        settings.args.sound.style = "dropdown"
        settings.args.sound.dialogControl = "LSM30_Sound"
        settings.args.sound.width = "double"
        settings.args.sound.order = 5.2

        settings.args.newOutput = settings.args.newOutput or {}
        settings.args.newOutput.name = L["Create new output"]
        settings.args.newOutput.type = "input"
        settings.args.newOutput.desc = L["Add an other output. Name may only consist letters, numbers, and spacings."]
        settings.args.newOutput.arg = arg
        settings.args.newOutput.validate = "ValidateOutputName"
        settings.args.newOutput.set = "CreateNewOutput"
        settings.args.newOutput.get = "GetNewOutput"
        settings.args.newOutput.order = 6

        local i = 1
        for key, value in pairs(Addon.db.profile.objectIconTypeOfInterest[objectIcon][name].sinks) do
            if value ~= false then

                local isDefault = not Addon.db.profile.objectIconTypeOfInterest[objectIcon][name].sinks[key].touched

                settings.args[key.."_sink"] = ns.sinkTable[arg..ns.SPLITKEY..key]:GetSinkAce3OptionsDataTable()
                if key == "BuildInDefaultOutputSink" then
                    settings.args[key.."_sink"].name = L["Default output"]
                else
                    settings.args[key.."_sink"].name = key
                end

                settings.args[key.."_sink"].order = 10+i/10
                settings.args[key.."_sink"].plugins = settings.args[key.."_sink"].plugins or {}
                settings.args[key.."_sink"].plugins[ADDONNAME] = settings.args[key.."_sink"].plugins[ADDONNAME] or {}
                settings.args[key.."_sink"].plugins[ADDONNAME].delete = settings.args[key.."_sink"].plugins[ADDONNAME].delete or {}
                settings.args[key.."_sink"].plugins[ADDONNAME].delete.name = L["Delete output"]
                settings.args[key.."_sink"].plugins[ADDONNAME].delete.type = "execute"
                settings.args[key.."_sink"].plugins[ADDONNAME].delete.desc = L["Click here to delete this output."]
                settings.args[key.."_sink"].plugins[ADDONNAME].delete.disabled = isDefault
                settings.args[key.."_sink"].plugins[ADDONNAME].delete.arg = arg..ns.SPLITKEY..key.."_sink"
                settings.args[key.."_sink"].plugins[ADDONNAME].delete.confirm = true
                settings.args[key.."_sink"].plugins[ADDONNAME].delete.confirmText = L["Delete %q output"]:format(settings.args[key.."_sink"].name)
                settings.args[key.."_sink"].plugins[ADDONNAME].delete.func = "DeleteSink"
                settings.args[key.."_sink"].plugins[ADDONNAME].delete.order = -0.1

                if isDefault then
                    settings.args[key.."_sink"].plugins[ADDONNAME].isDefault = settings.args[key.."_sink"].plugins[ADDONNAME].isDefault or {}
                    settings.args[key.."_sink"].plugins[ADDONNAME].isDefault.name = L["|cFF999999You can't delete a default output.|r"]
                    settings.args[key.."_sink"].plugins[ADDONNAME].isDefault.type = "description"
                    settings.args[key.."_sink"].plugins[ADDONNAME].isDefault.fontSize = "small"
                    settings.args[key.."_sink"].plugins[ADDONNAME].isDefault.width = "full"
                    settings.args[key.."_sink"].plugins[ADDONNAME].isDefault.order = -0.01
                end

                i = i + 1
            end
        end

        return settings
    end

    function Addon.GetDefaultsOptionTable(uiType, uiName, appName)
        defaultsOptions = defaultsOptions or {}
        defaultsOptions.name = L["Default Settings"]
        defaultsOptions.type = "group"
        defaultsOptions.desc = L["Default Settings for unchanged announcements."]
        defaultsOptions.handler = OptionsMethods
        defaultsOptions.set = "SetDefaults"
        defaultsOptions.get = "GetDefaults"
        defaultsOptions.args = defaultsOptions.args or {}

        defaultsOptions.args.active = defaultsOptions.args.active or {}
        defaultsOptions.args.active.name = L["Active"]
        defaultsOptions.args.active.type = "toggle"
        defaultsOptions.args.active.desc = L["If a new announcement should be active."]
        defaultsOptions.args.active.width = "double"
        defaultsOptions.args.active.order = 0

        defaultsOptions.args.showIconBefore = defaultsOptions.args.showIconBefore or {}
        defaultsOptions.args.showIconBefore.name = L["Show an icon before the announcement"]
        defaultsOptions.args.showIconBefore.type = "toggle"
        defaultsOptions.args.showIconBefore.desc = L["If a new announcement should show an icon before the announcement."]
        defaultsOptions.args.showIconBefore.width = "full"
        defaultsOptions.args.showIconBefore.order = 3.1

        defaultsOptions.args.showIconAfter = defaultsOptions.args.showIconAfter or {}
        defaultsOptions.args.showIconAfter.name = L["Show an icon after the announcement"]
        defaultsOptions.args.showIconAfter.type = "toggle"
        defaultsOptions.args.showIconAfter.desc = L["If a new announcement should show an icon after the announcement."]
        defaultsOptions.args.showIconAfter.width = "full"
        defaultsOptions.args.showIconAfter.order = 3.2

        defaultsOptions.args.iconSize = defaultsOptions.args.iconSize or {}
        defaultsOptions.args.iconSize.name = L["Icon size"]
        defaultsOptions.args.iconSize.type = "range"
        defaultsOptions.args.iconSize.desc = L["Size of the icon from a new announcement."]
        defaultsOptions.args.iconSize.min = 8
        defaultsOptions.args.iconSize.max = 512
        defaultsOptions.args.iconSize.step = 1
        defaultsOptions.args.iconSize.bigStep = 4
        defaultsOptions.args.iconSize.softMax = 64
        defaultsOptions.args.iconSize.order = 2

        defaultsOptions.args.mainTextColour = defaultsOptions.args.mainTextColour or {}
        defaultsOptions.args.mainTextColour.name = L["Main text colour"]
        defaultsOptions.args.mainTextColour.type = "color"
        defaultsOptions.args.mainTextColour.desc = L["Main text colour of a new announcement"]
        defaultsOptions.args.mainTextColour.hasAlpha = true
        defaultsOptions.args.mainTextColour.set = "SetDefaultColour"
        defaultsOptions.args.mainTextColour.get = "GetDefaultColour"
        defaultsOptions.args.mainTextColour.width = "full"
        defaultsOptions.args.mainTextColour.order = 4.1

        defaultsOptions.args.objectTextColour = defaultsOptions.args.objectTextColour or {}
        defaultsOptions.args.objectTextColour.name = L["Vignette text colour"]
        defaultsOptions.args.objectTextColour.type = "color"
        defaultsOptions.args.objectTextColour.desc = L["Vignette text colour of a new announcement"]
        defaultsOptions.args.objectTextColour.hasAlpha = true
        defaultsOptions.args.objectTextColour.set = "SetDefaultColour"
        defaultsOptions.args.objectTextColour.get = "GetDefaultColour"
        defaultsOptions.args.objectTextColour.width = "full"
        defaultsOptions.args.objectTextColour.order = 4.2

        defaultsOptions.args.playSound = defaultsOptions.args.playSound or {}
        defaultsOptions.args.playSound.name = L["Play sound"]
        defaultsOptions.args.playSound.type = "toggle"
        defaultsOptions.args.playSound.desc = L["If a new announcement should play a sound."]
        defaultsOptions.args.playSound.order = 5.1

        defaultsOptions.args.sound = defaultsOptions.args.sound or {}
        defaultsOptions.args.sound.name = L["Sound"]
        defaultsOptions.args.sound.type = "select"
        defaultsOptions.args.sound.desc = L["Which sound should be played."]
        defaultsOptions.args.sound.values = LSM:HashTable(LSM.MediaType.SOUND)
        defaultsOptions.args.sound.style = "dropdown"
        defaultsOptions.args.sound.dialogControl = "LSM30_Sound"
        defaultsOptions.args.sound.width = "double"
        defaultsOptions.args.sound.order = 5.2

        defaultsOptions.args.newOutput = defaultsOptions.args.newOutput or {}
        defaultsOptions.args.newOutput.name = L["Create new default output"]
        defaultsOptions.args.newOutput.type = "input"
        defaultsOptions.args.newOutput.desc = L["Add an other default output. Name may only consist letters, numbers, and spacings."]
        defaultsOptions.args.newOutput.validate = "ValidateDefaultOutputName"
        defaultsOptions.args.newOutput.set = "CreateNewDefaultOutput"
        defaultsOptions.args.newOutput.get = "GetNewOutput"
        defaultsOptions.args.newOutput.order = 6


        for key in pairs(MVAnnouncerDefaults.sinks) do
            if tmpDefaults.sinks[key] == nil then
                tmpDefaults.sinks[key] = {}
                for sinkKey, value in pairs(MVAnnouncerDefaults.sinks[key]) do
                    tmpDefaults.sinks[key][sinkKey] = value
                end
            end
        end
        local i = 1
        for key, value in pairs(tmpDefaults.sinks) do
            if value ~= false then
                defaultsOptions.args[key.."_sink"] = tmpDefaultsSinksTable[key]:GetSinkAce3OptionsDataTable()

                if key ~= '**' then
                    if key == "BuildInDefaultOutputSink" then
                        defaultsOptions.args[key.."_sink"].name = L["Default output"]
                    else
                        defaultsOptions.args[key.."_sink"].name = key
                    end

                    defaultsOptions.args[key.."_sink"].plugins = defaultsOptions.args[key.."_sink"].plugins or {}
                    defaultsOptions.args[key.."_sink"].plugins[ADDONNAME] = defaultsOptions.args[key.."_sink"].plugins[ADDONNAME] or {}
                    defaultsOptions.args[key.."_sink"].plugins[ADDONNAME].delete = defaultsOptions.args[key.."_sink"].plugins[ADDONNAME].delete or {}
                    defaultsOptions.args[key.."_sink"].plugins[ADDONNAME].delete.name = L["Delete output"]
                    defaultsOptions.args[key.."_sink"].plugins[ADDONNAME].delete.type = "execute"
                    defaultsOptions.args[key.."_sink"].plugins[ADDONNAME].delete.desc = L["Click here to delete this output."]
                    defaultsOptions.args[key.."_sink"].plugins[ADDONNAME].delete.arg = key
                    defaultsOptions.args[key.."_sink"].plugins[ADDONNAME].delete.confirm = true
                    defaultsOptions.args[key.."_sink"].plugins[ADDONNAME].delete.confirmText = L["Delete %q output"]:format(defaultsOptions.args[key.."_sink"].name)
                    defaultsOptions.args[key.."_sink"].plugins[ADDONNAME].delete.func = "DeleteDefaultSink"
                    defaultsOptions.args[key.."_sink"].plugins[ADDONNAME].delete.order = -0.1
                else
                    defaultsOptions.args[key.."_sink"].name = L["Standard"]
                    defaultsOptions.args[key.."_sink"].desc = L["This are the settings for all not changed and new created outputs."]
                    defaultsOptions.args[key.."_sink"].plugins = defaultsOptions.args[key.."_sink"].plugins or {}
                    defaultsOptions.args[key.."_sink"].plugins[ADDONNAME] = defaultsOptions.args[key.."_sink"].plugins[ADDONNAME] or {}
                    defaultsOptions.args[key.."_sink"].plugins[ADDONNAME].title = defaultsOptions.args[key.."_sink"].plugins[ADDONNAME].title or {}
                    defaultsOptions.args[key.."_sink"].plugins[ADDONNAME].title.name = L["This are the settings for all not changed and new created outputs."]
                    defaultsOptions.args[key.."_sink"].plugins[ADDONNAME].title.type = "description"
                    defaultsOptions.args[key.."_sink"].plugins[ADDONNAME].title.fontSize = "medium"
                    defaultsOptions.args[key.."_sink"].plugins[ADDONNAME].title.order = 1
                end

                defaultsOptions.args[key.."_sink"].order = 10+i/10
                i = i + 1
            end
        end

        defaultsOptions.args.applyDefaults = defaultsOptions.args.applyDefaults or {}
        defaultsOptions.args.applyDefaults.name = L["Apply defaults"]
        defaultsOptions.args.applyDefaults.type = "execute"
        defaultsOptions.args.applyDefaults.desc = L["Set this settings as defaults."]
        defaultsOptions.args.applyDefaults.func = "ApplyDefaults"
        defaultsOptions.args.applyDefaults.order = 0

        return defaultsOptions
    end

    function Addon:CleanOptions()
        wipe(announcementOptions)
        wipe(options)
        wipe(ns.sinkTable)
        self:NotifyOptionsChange()
    end

    function Addon:RemoveVignette(objectIcon, objectName)
        if type(options.args[objectIcon]) == "table" then
            options.args[objectIcon].args[objectName] = nil
        end
        self:NotifyOptionsChange()
    end

    function Addon:NotifyOptionsChange()
        LibStub("AceConfigRegistry-3.0"):NotifyChange(ADDONNAME)
    end

    function Addon:RegisterOptions()
        local AceConfig = LibStub("AceConfig-3.0")
        local AceConfigDialog = LibStub("AceConfigDialog-3.0")

        AceConfig:RegisterOptionsTable(ADDONNAME, self.GetOptionsTable )
        AceConfig:RegisterOptionsTable(ADDONNAME.."defaults", self.GetDefaultsOptionTable )
        AceConfig:RegisterOptionsTable(ADDONNAME.."profile", LibStub("AceDBOptions-3.0"):GetOptionsTable(self.db) )

        AceConfigDialog:AddToBlizOptions(ADDONNAME, ADDONNAME)
        AceConfigDialog:AddToBlizOptions(ADDONNAME.."defaults", L["Defaults"], ADDONNAME)
        AceConfigDialog:AddToBlizOptions(ADDONNAME.."profile", L["Profile"], ADDONNAME)
    end
end







